/*
 * ml.h
 *
 *
 * Part of TREE-PUZZLE 5.2 (July 2004)
 *
 * (c) 2003-2004 by Heiko A. Schmidt, Korbinian Strimmer, and Arndt von Haeseler
 * (c) 1999-2003 by Heiko A. Schmidt, Korbinian Strimmer,
 *                  M. Vingron, and Arndt von Haeseler
 * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed under
 * the GNU public licence.  See http://www.opensource.org for details.
 *
 *
 */


/* values for mlmode */
#define ML_QUART 1
#define ML_TREE  2
#define NJ_TREE  3


/* values for rhetmode */
#define UNIFORMRATE 0
#define GAMMARATE   1
#define TWORATE     2
#define MIXEDRATE   3


/* definitions */

#define MINTS 0.20  /* Ts/Tv parameter */
#define MAXTS 30.0
#define MINYR 0.10  /* Y/R Ts parameter */
#define MAXYR 6.00
#define MINFI 0.00  /* fraction invariable sites */
#define MAXFI 0.99  /* only for input */
#define MINGE 0.01  /* rate heterogeneity parameter */
#define MAXGE 0.99
#define MINCAT 4    /* discrete Gamma categories */
#define MAXCAT 16

#define RMHROOT         5.0     /* upper relative bound for height of root   */    
#define MAXARC          900.0   /* upper limit on branch length (PAM) = 6.0  */
#define MINARC          0.001   /* lower limit on branch length (PAM) = 0.00001 */

#ifdef USE_ADJUSTABLE_EPS
    /* error in branch length (PAM) = 0.000001   */
#   define EPSILON_BRANCH_DEFAULT  0.0001  
#   define EPSILON_BRANCH epsilon_branch
     double epsilon_branch;
#else
#   define EPSILON_BRANCH  0.0001  
#endif

#ifdef USE_ADJUSTABLE_EPS
/* error in node and root heights            */
#   define EPSILON_HEIGHTS_DEFAULT 0.0001
#   define EPSILON_HEIGHTS epsilon_heights
     double epsilon_heights;
#else
#   define EPSILON_HEIGHTS 0.0001
#endif

#ifdef USE_ADJUSTABLE_EPS
	double epsilon_branch = EPSILON_BRANCH_DEFAULT;
	double epsilon_heights = EPSILON_HEIGHTS_DEFAULT;
#endif

#define MAXIT           100     /* maximum number of iterates of smoothing   */
#define MINFDIFF        0.00002 /* lower limit on base frequency differences */
#define MINFREQ         0.0001  /* lower limit on base frequencies = 0.01%   */
#define NUMQBRNCH       5       /* number of branches in a quartet           */
#define NUMQIBRNCH      1       /* number of internal branches in a quartet  */
#define NUMQSPC         4       /* number of sequences in a quartet          */

/* 2D minimisation */

#ifdef USE_ADJUSTABLE_EPS
    /* epsilon substitution process estimation */
#   define EPSILON_SUBSTPARAM_DEFAULT 0.01
#   define EPSILON_SUBSTPARAM epsilon_substparam 
     double epsilon_substparam;
#else
#   define EPSILON_SUBSTPARAM 0.01
#endif

#ifdef USE_ADJUSTABLE_EPS
    /* epsilon rate heterogeneity estimation  */
#   define EPSILON_RATEPARAM_DEFAULT  0.01
#   define EPSILON_RATEPARAM  epsilon_rateparam
     double epsilon_rateparam;
#else
#   define EPSILON_RATEPARAM 0.01
#endif

#ifdef USE_ADJUSTABLE_EPS
	double epsilon_substparam = EPSILON_SUBSTPARAM_DEFAULT;
	double epsilon_rateparam = EPSILON_RATEPARAM_DEFAULT;
#endif

/* quartet series */
#define MINPERTAXUM 2
#define MAXPERTAXUM 6
#define TSDIFF 0.20
#define YRDIFF 0.10

/* type definitions */

typedef struct node
{
	struct node *isop;	/* pointers within node (nodes) */
	struct node *kinp;	/* pointers between nodes (branches) */
	int descen;
	int number;
	double length;		/* branch length inferred (no clock) */
	double lengthc;		/* branch length inferred (clock) */
	double lengthext;	/* external length from usertree */
	int    lengthset;	/* length set by usertree */
	double varlen;
	double height;
	double varheight;
	ivector paths;		/* path directions for Korbi's p-step */
	cvector eprob;
	dcube partials;		/* partial likelihoods */
	char *label;		/* internal labels */
} Node;

typedef struct tree
{
	Node    *rootp;
	Node   **ebrnchp;	/* list of pointers to external branches     */
	Node   **ibrnchp;       /* list of pointers to internal branches     */
	double   lklhd;	        /* total log-likelihood                      */
	double   lklhdc;        /* total log-likelihood clock                */
	dmatrix  condlkl;	/* lhs for each pattern and non-zero rate    */
	double   rssleast;
} Tree;


/* global variables */

 Node    *chep;      /* pointer to current height node                 */
 Node    *rootbr;    /* pointer to root branch                         */
 Node   **heights;   /* pointer to height nodes in unrooted tree       */
 int      Numhts;    /* number of height nodes in unrooted tree        */
 double   hroot;     /* height of root                                 */
 double   varhroot;  /* variance of height of root                     */
 double   maxhroot;  /* maximal height of root                         */
 int      locroot;   /* location of root                               */
 int      numbestroot; /* number of best locations for root            */
 int      clockmode; /* clocklike vs. nonclocklike computation         */
 cmatrix  Identif;   /* sequence names                                 */
 cmatrix  Namestr;   /* sequence names (delimited by '\0')             */
 cmatrix  Seqchar;   /* ML sequence data                               */
 cmatrix  Seqpat;    /* ordered site patterns                          */
 ivector  constpat;  /* indicates constant site patterns               */
 cvector  seqchi;
 cvector  seqchj;
 dcube    partiali;
 dcube    partialj;
 dcube    ltprobr;   /* transition probabilites (for all non-zero rates*/
 dmatrix  Distanmat;   /* matrix with maximum likelihood distances     */
 dmatrix  Evec;        /* Eigenvectors                                 */
 dmatrix  Ievc;        /* Inverse eigenvectors                         */
 double   TSparam;     /* Ts/Tv parameter                              */
 double   GTR_ACrate;  /* A <-> G mutation rate for GTR                */
 double   GTR_AGrate;  /* A <-> C mutation rate for GTR                */
 double   GTR_ATrate;  /* A <-> T mutation rate for GTR                */
 double   GTR_CGrate;  /* C <-> G mutation rate for GTR                */
 double   GTR_CTrate;  /* C <-> T mutation rate for GTR                */
 double   GTR_GTrate;  /* G <-> T mutation rate for GTR                */
 double   tsmean, yrmean;
 double   YRparam;     /* Y/R Ts parameter                             */
 double   geerr;       /* estimated error of rate heterogeneity        */
 double   Geta;        /* rate heterogeneity parameter                 */
 double   fracconst;   /* fraction of constant sites                   */
 double   fracconstpat;/* fraction of constant patterns                */
 double   Proportion;  /* for tree drawing                             */
 double   tserr;       /* estimated error of TSparam                   */
 double   yrerr;       /* estimated error of YRparam                   */
 double   fracinv;     /* fraction of invariable sites                 */
 double   fierr;       /* estimated error of fracinv                   */
 dvector  Brnlength;
 dvector  Distanvec;
 dvector  Eval;        /* Eigenvalues of 1 PAM rate matrix             */
 dvector  Freqtpm;     /* base frequencies                             */
 dvector  Rates;       /* rate of each of the categories               */
 dmatrix  iexp;
 imatrix  Basecomp;    /* base composition of each taxon               */
 ivector  usedtaxa;    /* list needed in the input treefile procedure  */
 int      numtc;       /* auxiliary variable for printing rooted tree  */
 int      qcalg_optn;  /* use quartet subsampling algorithm            */
 int      approxp_optn;/* approximate parameter estimation             */
 int      chi2fail;    /* flag for chi2 test                           */
 int      Converg;     /* flag for ML convergence (no clock)           */
 int      Convergc;    /* flag for ML convergence (clock)              */
 int      data_optn;   /* type of sequence input data                  */  
 int      Dayhf_optn;  /* Dayhoff model                                */
 int      HKY_optn;    /* use HKY model                                */
 int      Jtt_optn;    /* JTT model                                    */
 int      print_GTR_optn; /* print (restricted) GTR matrix */
 int      GTR_optn;    /* GTR model                                    */
 int      blosum62_optn; /* BLOSUM 62 model                            */
 int      mtrev_optn;  /* mtREV model                                  */
 int      cprev_optn;  /* cpREV model                                  */
 int      vtmv_optn;   /* VT model                                     */
 int      wag_optn;    /* WAG model                                    */
 int      Maxsite;     /* number of ML characters per taxum            */
 int      Maxspc;      /* number of sequences                          */
 int      mlmode;      /* quartet ML or user defined tree ML           */
 int      nuc_optn;    /* nucleotide (4x4) models                      */
 int      Numbrnch;    /* number of branches of current tree           */
 int      numcats;     /* number of rate categories                    */
 int      Numconst;    /* number of constant sites                     */
 int      Numconstpat; /* number of constant patterns                  */
 int      Numibrnch;   /* number of internal branches of current tree  */
 int      Numitc;      /* number of ML iterations assumning clock      */
 int      Numit;       /* number of ML iterations if convergence       */
 int      Numptrn;     /* number of site patterns                      */
 int      Numspc;      /* number of sequences of current tree          */
 int      optim_optn;  /* optimize model parameters                    */
 int      grate_optim; /* optimize Gamma rate heterogeneity parameter  */
 int      SH_optn;     /* SH nucleotide (16x16) model                  */
 int      TN_optn;     /* use TN model                                 */
 int      tpmradix;    /* number of different states                   */
 int      fracinv_optim;  /* optimize fraction of invariable sites     */
 int      typ_optn;    /* type of PUZZLE analysis                      */
 ivector  Weight;      /* weight of each site pattern                  */
 Tree    *Ctree;       /* pointer to current tree                      */
 int      qca, qcb, qcc, qcd; /* quartet currently optimized           */
 ivector  Alias;       /* link site -> corresponding site pattern      */
 ivector  bestrate;    /* optimal assignment of rates to sequence sites*/

 int      bestratefound;  /* best rate per site already reconstructed ? */

/* function prototypes of all ml function */

void convfreq(dvector);
void tp_radixsort(cmatrix, ivector, int, int, int *);
void condenceseq(cmatrix, ivector, cmatrix, ivector, int, int, int);
void countconstantsites(cmatrix, ivector, int, int, int *, int*);
void evaluateseqs(void);
void elmhes(dmatrix, ivector, int);
void eltran(dmatrix, dmatrix, ivector, int);
void mcdiv(double, double, double, double, double *, double *);
void hqr2(int, int, int, dmatrix, dmatrix, dvector, dvector);
void onepamratematrix(dmatrix);
void eigensystem(dvector, dmatrix);
void luinverse(dmatrix, dmatrix, int);
void checkevector(dmatrix, dmatrix, int);
void tranprobmat(void);
void tprobmtrx(double, dmatrix);
double comptotloglkl(dmatrix);
void allsitelkl(dmatrix, dvector);
void writesitelklmatrixphy(FILE *outf, int numut, int numsite, dmatrix allslkl, dmatrix allslklc, int clock);
void writesitelklvectorphy(FILE *outf, char *name, int numsite, dvector aslkl);
/* void writesitelklbin(FILE *, dvector); */ /* TODO */
double pairlkl(double);
double mldistance(int, int, double); /*epe*/
/* double mldistance(int, int); */
void initdistan(void);
void computedistan(void);
void productpartials(Node *);
void partialsinternal(Node *);
void partialsexternal(Node *);
void initpartials(Tree *);
double intlkl(double);
void optinternalbranch(Node *);
double extlkl(double);
void optexternalbranch(Node *);
void finishlkl(Node *);
double optlkl(Tree *);
double treelkl(Tree *);
void luequation(dmatrix, dvector, int);
void lslength(Tree *, dvector, int, int, dvector);
void setextlength(Tree *, int, int, dvector);
void getusertree(FILE *, cvector, int, int);
Node *internalnode(Tree *, char **, int *);
void constructtree(Tree *, cvector, int);
void removebasalbif(cvector);
void makeusertree(FILE *, int);
Tree *new_tree(int, int, cmatrix);
Tree *new_quartet(int, cmatrix);
void free_tree(Tree *, int);
void make_quartet(int, int, int, int);
void changedistan(dmatrix, dvector, int);
double quartet_lklhd(int, int, int, int);
double quartet_alklhd(int, int, int, int);
void readusertree(FILE *, int);
double usertree_lklhd(int, int);
double usertree_alklhd(int, int);
void mlstart(void);
void distupdate(int, int, int, int);
void mlfinish(void);
void prbranch(Node *, int, int, int, ivector, ivector, FILE *);
void getproportion(double *, dvector, int);
void prtopology(FILE *);
void fputphylogeny(FILE *);
void resulttree(FILE *outfp, int usebranch);
void njtree(FILE *);
void njdistantree(Tree *);
void findbestratecombination(void);
void printbestratecombination(FILE *);
void printbestratecombinationtofile(FILE *, int);
int checkedge(int);
void fputsubstree(FILE *, Node *);
void fputrooted(FILE *, int);
void findheights(Node *);
void initclock(int);
double clock_alklhd(int);
double heightlkl(double);
void optheight(void);
double rheightlkl(double);
void optrheight(void);
double clock_lklhd(int);
int findrootedge(void);
void resultheights(FILE *);

/* mlparam.c */
double homogentest(int);
void YangDiscreteGamma(double, int, double *);
void updaterates(void);
void computestat(double *, int, double *, double *);
double quartetml(int, int, int, int);
double opttsq(double);
double optyrq(double);
void optimseqevolparamsquart(void);
double opttst(double);
double optyrt(double);
void optimseqevolparamstree(void);
double optfi(double);
double optge(double);
void optimrateparams(void);

void estimateparametersnotree(void); /* moved from puzzle1.c */
void estimateparameterstree(void);   /* moved from puzzle1.c */
/* mlparam.c end */

int gettpmradix(void);
void rtfdata(dmatrix, double *);
int code2int(cvector, int *, int *);
char *int2code(int);

void jttdata(dmatrix, double *);
void dyhfdata(dmatrix, double *);
void mtrevdata(dmatrix, double *);
void cprev45data(dmatrix, double *);
void blosum62data(dmatrix, double *);
void vtmvdata(dmatrix, double *);
void wagdata(dmatrix, double *);

